import os


def find_path(paths):
    for path in paths:
        try:
            assert type(path) is str
            if os.path.exists(path):
                return path
        except Exception as e:
            raise e
    return None



import os
from visdom import Visdom


def visdom():
    # check visdom connection
    viz = Visdom(server='http://127.0.0.1', port=8097)
    if not viz.check_connection(timeout_seconds=5):
        os.system('nohup python -m visdom.server &')
    assert viz.check_connection(timeout_seconds=5)
    return viz


# from PIL import Image
# from torchvision import transforms

# viz = visdom_server()
# viz.image(
#     transforms.ToTensor()(Image.open('/home/sh/test.jpg').convert('RGB')),
#     opts={
#         'title': 'TestHakase',
#         'showlegend': True
#     }
# )



from functools import wraps


def raise_err(func):
    @wraps(func)
    def core(*args, **kwargs):
        out, err = func(*args, **kwargs)
        if len(err) > 0:
            raise err[0]
    return core


def try_callback(func):
    @wraps(func)
    def core(*args, **kwargs):
        try:
            func(*args, **kwargs)
            return [], []
        except Exception as e:
            return [], [e]
    return core



import torch


def step(model: torch.nn.Module,
         x: torch.Tensor,
         target: torch.Tensor,
         criterion,
         optimizer: torch.optim.Optimizer,
         **kwargs):
    assert model.training
    optimizer.zero_grad()
    y = model(x)
    loss = criterion(y, target, **kwargs)
    loss.backward()
    optimizer.step()
    return loss.item()


def inference(model: torch.nn.Module,
              x: torch.Tensor,
              target: torch.Tensor,
              criterion,
              **kwargs):
    assert not model.training
    y = model(x)
    loss = criterion(y, target, **kwargs)
    return loss.item()





import os
from functools import wraps
import paramiko
import base64

from .utils import try_callback, raise_err


class Tunnel():
    _status_head = (
        '          ',
        '[ \033[1;32m  ok  \033[0m ]',
        '[ \033[1;31mfailed\033[0m ]',
    )

    def __init__(self, key, mode='verbose'):
        self.__key = key
        self.__client = None
        self.__root = '~'
        _msg_levels = ('n', 'c', 'e', 'v')  # nothing < command < error < verbose
        assert mode in _msg_levels
        self._output_level = _msg_levels.index(mode.lower()[0])

    @staticmethod
    def decode(key: str):
        sequence = base64.b64decode(key)
        return str(sequence, 'utf8').split('$')

    @staticmethod
    def connect(key):
        ssh = paramiko.SSHClient()  # 创建sshclient
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())  # 指定当对方主机没有本机公钥的情况时应该怎么办，AutoAddPolicy表示自动在对方主机保存下本机的秘钥
        ssh.connect(*Tunnel.decode(key))
        return ssh

    @staticmethod
    def sftp_callback(transferred, total):
        print(f'\r{Tunnel._status_head[0]} transferred/total: {transferred}/{total} ({transferred/total:%})', end='')

    @staticmethod
    def ssh_callback(level, line=None, status=None, root=None, refresh_status=None):
        if level > 0:
            if refresh_status is not None:
                print('\r'+Tunnel._status_head[refresh_status])
            elif status is not None:
                print(f'{status} {root}$ {line}', end='')
            else:
                print(f'{Tunnel._status_head[0]} {line}', end='')

    def callback(func):
        @wraps(func)
        def core(self, command):
            # ('nothing', 'command', 'error', 'verbose') = 0, 1, 2, 3
            if command.startswith('mv') or command.startswith('cp'):
                Tunnel.ssh_callback(self._output_level, command+'\n', self._status_head[1], self.__root)  # sftp related command returns ok as default
                out, err = func(self, command)
                print('\n', end='')
                return out, err
            Tunnel.ssh_callback(self._output_level, command, self._status_head[0], self.__root)
            out, err = func(self, command)
            Tunnel.ssh_callback(self._output_level, refresh_status=(2 if len(err) > 0 else 1))
            for line in err:
                Tunnel.ssh_callback(self._output_level-1, str(line))
            for line in out:
                Tunnel.ssh_callback(self._output_level-2, line)
            return out, err
        return core

    def __enter__(self):
        self.__client = self.connect(self.__key)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.__client.close()

    def __call__(self, command):
        _, stdout, stderr = self.__client.exec_command(f'. /etc/profile;. .bashrc;cd {self.__root};{command}')
        out = stdout.readlines()
        err = stderr.readlines()
        return out, err

    @raise_err
    @try_callback
    def __upload__(self, local, remote):
        with self.__client.open_sftp() as sftp:
            sftp.put(local, remote, callback=self.sftp_callback)

    @raise_err
    @try_callback
    def __download__(self, remote, local):
        with self.__client.open_sftp() as sftp:
            sftp.get(remote, local, callback=self.sftp_callback)

    @raise_err
    @try_callback
    def __nativecall__(self, command):
        os.system(command)

    @callback
    def call(self, command):
        out, err = [], []
        # supports for 'cd' command, allows relocating before each call.
        if command.startswith('cd'):
            out, err = self.__call__(command)
            if len(err) == 0:
                self.__root = command.split(' ')[-1]
                self.__root = self.__call__('pwd')[0][0].replace('\n', '')
            return out, err
        # hyper supports for 'mv' command, allows 'local:' pattern for sftp.
        if command.startswith('mv') or command.startswith('cp'):
            _, a, b = command.split(' ')
            if a.startswith('local:'):
                a = a.split('local:')[-1]
                b = self.__root + '/' + b
                self.__upload__(a, b)
                if command.startswith('mv'):
                    self.__nativecall__(f'rm {a}')
            elif b.startswith('local:'):
                b = b.split('local:')[-1]
                a = self.__root + '/' + a
                self.__download__(a, b)
                if command.startswith('mv'):
                    out, err = self.__call__(f'rm {a}')
            else:
                out, err = self.__call__(f'mv {a} {b}')
            return out, err
        return self.__call__(command)

    def cd(self, dir):
        return self.call('cd ' + dir)

    def tarc(self, src, target):
        return self.call('tar -jcvf ' + src + ' ' + target)

    def tarx(self, src, target):
        return self.call('tar -jxvf ' + src + ' -C ' + target)

    def mv(self, src, target):
        return self.call('mv ' + src + ' ' + target)

    def cp(self, src, target):
        return self.call('cp ' + src + ' ' + target)


def keygen(ip_address, user, password, port=22):
    sequence = f'{ip_address}${port}${user}${password}'
    return str(base64.b64encode(sequence.encode('utf8')), 'utf8')


def open(key, mode='v'):
    return Tunnel(key, mode)



import os


from .utils import bz2, try_callback, raise_err
from . import ssh


def device(key, mode='v'):
    return Device(key, mode)


class Device():

    def __init__(self, key, mode='verbose'):
        self.__key = key
        self.__tid = None
        _msg_levels = ('n', 'c', 'e', 'v')
        assert mode in _msg_levels
        self._output_level = _msg_levels.index(mode)
        self.local_root = './clones'
        self.device_root = '~/clones'

    def __enter__(self):
        if not os.path.exists(self.local_root):
            os.makedirs(self.local_root)
        with self.__handler__('n') as f:
            f.call(f'mkdir {self.device_root}')
            f.cd(self.device_root)
            f.call(f'mkdir todo;mkdir env')
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def __handler__(self, mode):
        return ssh.open(self.__key, mode=mode)

    def call(self, command, mode='v'):
        with ssh.open(self.__key, mode=mode) as f:
            _, err_ = f.cd(self.device_root)
            out, err = f.call(command)
        return out, err_ + err

    def commit(self, src, tid=None, ignore=r'.*\.(pdf|dio|png|gif)'):
        return bz2.compress(src, self.local_root, ignore, tid)

    def checkout(self, tid):
        self.__tid = tid
        return tid

    @raise_err
    @try_callback
    def fetch(self, tid=None):
        if tid is None:
            tid = self.__tid
        assert tid is not None
        with self.__handler__('e') as f:
            f.cd(self.device_root + '/env')
            f.tarc(tid + '.tar.bz2', tid)
            f.mv(tid + '.tar.bz2', f'local:{self.local_root}/{tid}.tar.bz2')
        os.system(f'tar -jxvf {self.__root}/{tid}.tar.bz2 -C {self.__root}; rm {self.__root}/{tid}.tar.bz2')

    @raise_err
    @try_callback
    def push(self):
        with self.__handler__('e') as f:
            f.cd(self.device_root + '/todo')
            for branch in os.listdir(self.local_root):
                if branch.endswith('.tar.bz2'):
                    f.cp(f'local:{self.local_root}/{branch}', branch)
        os.system(f'rm -r {self.local_root}/*.tar.bz2')

    @raise_err
    @try_callback
    def pull(self):
        with self.__handler__('e') as f:
            f.cd(self.device_root)
            f.tarc('env.tar.bz2', './env')
            f.mv('env.tar.bz2', f'local:{self.local_root}/env.tar.bz2')
            f.call(f'rm ./env/* -r')
        os.system(f'tar -jxvf {self.local_root}/env.tar.bz2 -C {self.local_root};')
        os.system(f'rm {self.local_root}/env.tar.bz2; mv {self.local_root}/env/* {self.local_root}/; rm {self.local_root}/env -r')

    @raise_err
    @try_callback
    def ls(self, tid=None, pipe=None):
        if tid is None:
            tid = self.__tid
        assert tid is not None
        with self.__handler__('v') as f:
            f.cd(self.device_root + '/env')
            f.call(f'cat {tid}/log'+('' if pipe is None else f'|{pipe}'))

    @raise_err
    @try_callback
    def run(self, script):
        with self.__handler__('v') as f:
            f.cd(self.device_root + '/env')
            f.cp('local:'+script, './god.py')
            f.call(f'nohup python -u god.py >log 2>&1 &')



#!/usr/bin/env python
# -*- coding:utf-8 -*-


from setuptools import setup, find_packages
import getopt
import sys
import os

if __name__ == "__main__":

    # python setup.py sdist
    # twine upload dist/torchmods-{version}.tar.gz'

    # set version
    version = '0.4.2'

    # upload
    setup(
        name="torchmods",
        version=version,
        keywords=("cv", "pytorch", "auxiliary"),
        description="mods for torch & cv",
        long_description="experimental mods for pytorch & cv research",
        license="MIT Licence",

        url="https://github.com/klrc/torchmods",
        author="klrc",
        author_email="sh@mail.ecust.edu.com",

        packages=find_packages(),
        include_package_data=True,
        platforms=["all"],
        install_requires=["torch",
                            "matplotlib",
                            "paramiko",
                            "opencv_python",
                            "requests",
                            "numpy",
                            "torchvision",
                            "beautifulsoup4",
                            "imageio",
                            "Pillow",
                            "pynvml"]
    )
    print('to install:\n  pip install torchmods -U -i https://pypi.org/simple')


from typing import Union, Tuple
import os
import imageio
import numpy as np
import cv2
import torch
import matplotlib.pyplot as plt


class NumpyImage():
    def __init__(self, x, imtype=np.uint8):
        self.data = x.astype(imtype)

    def save(self, output):
        plt.imshow(self.data)
        plt.savefig(output)
        plt.close()
        return self

    def __add__(self, other):
        if isinstance(other, NumpyImage):
            other = other.data
        return NumpyImage(self.data + other)

    def __mul__(self, other):
        if isinstance(other, NumpyImage):
            other = other.data
        return NumpyImage(self.data * other)


def denormalize(x):
    """"将tensor的数据类型转成numpy类型，并反归一化.

    Parameters:
        input_image (tensor) --  输入的图像tensor数组
        imtype (type)        --  转换后的numpy的数据类型
    """
    mean = [0.485, 0.456, 0.406]
    std = [0.229, 0.224, 0.225]

    if not isinstance(x, np.ndarray):
        if isinstance(x, torch.Tensor):  # get the data from a variable
            if x.shape[0] == 1 and len(x.shape) == 4:
                x = x[0]
            image_tensor = x.data
        else:
            return x
        image_numpy = image_tensor.cpu().float().numpy()  # convert it into a numpy array
        if image_numpy.shape[0] == 1:  # grayscale to RGB
            image_numpy = np.tile(image_numpy, (3, 1, 1))
        for i in range(len(mean)):
            image_numpy[i] = image_numpy[i] * std[i] + mean[i]
        image_numpy[image_numpy < 0] = 0
        image_numpy[image_numpy > 1] = 1
        image_numpy = image_numpy * 255
        image_numpy = np.transpose(image_numpy, (1, 2, 0))  # post-processing: tranpose and scaling
    else:  # if it is a numpy array, do nothing
        image_numpy = x
    return NumpyImage(image_numpy)


def heatmap(x):
    if len(x.shape) == 4:
        x = x[0]
    assert x.shape[0] == 1
    heatmap = x[0].cpu().data.numpy()
    heatmap = heatmap/np.max(heatmap)
    # must convert to type unit8
    heatmap = np.uint8(255 * heatmap)
    heatmap = cv2.applyColorMap(heatmap, cv2.COLORMAP_JET)
    return NumpyImage(heatmap)


class Gif():
    def __init__(self, dir, sort_func=lambda x: x, reverse=False):
        self.data = self.load(dir, sort_func, reverse)

    @staticmethod
    def load(dir, sort_func=lambda x: x, reverse=False):
        files = [x for x in os.listdir(dir) if x.endswith('jpg') or x.endswith('png')]
        files.sort(key=sort_func, reverse=reverse)
        imgs = [imageio.imread(f'{dir}/{img_file}') for img_file in files]
        return imgs

    def save(self, output, fps=8):
        imageio.mimsave(output, self.data, fps=fps)
        return self


class ImageGrid():
    def __init__(self, shape: Union[int, Tuple[int, int]]):
        self.data = []
        if type(shape) is not Tuple:
            shape = (1, shape)
        self.shape = shape
        self._maxlen = shape[0]*shape[1]

    def push(self, x):
        if len(self.data) >= self._maxlen:
            raise OverflowError
        if isinstance(x, NumpyImage):
            x = x.data
        self.data.append(x)
        return self

    def save(self, output, axis_off=True, title=None, quality=4):
        # reshape data
        lines, rows = self.shape
        plt.figure(figsize=(rows*quality, lines*quality))
        for i, x in enumerate(self.data):
            if axis_off:
                plt.axis('off')
            subp = plt.subplot(lines, rows, i+1)
            if title:
                subp.set_title(title)
            plt.imshow(x)
        if axis_off:
            plt.axis('off')
        plt.savefig(output)
        plt.close()
        return self

    def clear(self):
        self.data = []
        return self




import os

import cv2
import numpy as np
import torch
import torchmods.nn as nn
import torchvision.transforms as transforms
from torchmods.nn.atomic import inference, step
from torchmods.utils import find_path, nvidia, printlog

from .dataset_collection import dataloaders, datasets


def padding(img, shape_r=240, shape_c=320, channels=3):
    img_padded = np.zeros((shape_r, shape_c, channels), dtype=np.uint8)
    if channels == 1:
        img_padded = np.zeros((shape_r, shape_c), dtype=np.uint8)

    original_shape = img.shape
    rows_rate = original_shape[0]/shape_r
    cols_rate = original_shape[1]/shape_c

    if rows_rate > cols_rate:
        new_cols = (original_shape[1] * shape_r) // original_shape[0]
        img = cv2.resize(img, (new_cols, shape_r))
        if new_cols > shape_c:
            new_cols = shape_c
        img_padded[:, ((img_padded.shape[1] - new_cols) // 2):((img_padded.shape[1] - new_cols) // 2 + new_cols)] = img
    else:
        new_rows = (original_shape[0] * shape_c) // original_shape[1]
        img = cv2.resize(img, (shape_c, new_rows))
        if new_rows > shape_r:
            new_rows = shape_r
        img_padded[((img_padded.shape[0] - new_rows) // 2):((img_padded.shape[0] - new_rows) // 2 + new_rows), :] = img

    return img_padded


def preprocess_maps(path, shape_r=240, shape_c=320):
    ims = np.zeros((1, shape_r, shape_c))

    original_map = cv2.imread(path, 0)
    padded_map = padding(original_map, shape_r, shape_c, 1)
    ims[0] = padded_map.astype(np.float32)
    ims[0] /= 255.0

    return ims.astype(np.float32)


class SALICON(torch.utils.data.Dataset):
    img_dir = 'images'
    seed = 0

    def __init__(self, root, train=True, transform=None):
        super().__init__()
        self.load_data(root, train)
        self.transform = transform
        self.lb_type = 'maps'

    def __cvimg__(self, path, convert2rgb=True):
        """
            read image in RGB mode.
        """
        img_cv = cv2.imread(path)
        if convert2rgb:
            img_cv = cv2.cvtColor(img_cv, cv2.COLOR_BGR2RGB)
        return img_cv

    def __getitem__(self, index):
        """
            function for torch.util.data.Dataset class,
            returns image, (target_1, target_2, ..., target_n)

            Args:
                index:
        """
        img = self.__cvimg__(self.data[index])
        if self.transform:
            img = self.transform(img)
        if self.lb_type == 'maps':
            target = preprocess_maps(self.data[index].replace('/images/', '/maps/').replace('.jpg', '.png'))
            target = torch.from_numpy(target)
            shape = target.shape
            sum_target = torch.nn.functional.adaptive_avg_pool2d(target, output_size=(1, 1))*shape[-2]*shape[-1]
            target /= sum_target
        return img, target

    def __len__(self):
        """
            function for torch.util.data.Dataset class.
        """
        return len(self.data)

    def load_data(self, root, train):
        """
            pre-load image path,
            (override if more complex func needed),
            Args:
                root:
                train:
        """
        assert os.path.exists(root)
        src_dir = 'train' if train else 'val'
        src_dir = f'{root}/images/{src_dir}'
        self.data = [f'{src_dir}/{x}' for x in os.listdir(src_dir)]


class EnvSALICON():
    def __init__(self):
        self.root = find_path(['/home/sh/datasets/SALICON', '/home/public/datasets/SALICON'])
        self.transform = transforms.Compose([
            transforms.ToPILImage(),
            transforms.Resize(size=(240, 320)),
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        ])
        self.max_training_patience = 8
        self.checkpoints = torch.linspace(0, 1, steps=20)
        self.max_epoch = 100
        self.batch_size = 16
        self.num_workers = 2
        self.pin_memory = True

    def sample(self, batch_size=1, shuffle=True):
        from torch.utils.data import DataLoader
        for batch in DataLoader(SALICON(root=self.root, train=False, transform=self.transform), batch_size=batch_size, shuffle=shuffle):
            return batch

    def check(self, model, optimizer):
        model_size = nvidia.state()['used']
        torch.autograd.set_detect_anomaly(True)
        self.train(model, optimizer, _test_mode=True)
        self.performance(model, _test_mode=True)
        print(f'> model size: {model_size}.')
        print('> passed.')

    def train(self, model, optimizer, _test_mode=False):
        trainloader, valloader, testloader = dataloaders(
            datasets(SALICON, root=self.root, transforms=(self.transform,)*3),
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )
        criterion = nn.LogSoftmaxWrapper(nn.KLDivLoss(reduction='batchmean'))
        min_average_kldiv_test = float('inf')
        patience = self.max_training_patience
        checkpoints = (self.checkpoints*len(trainloader)).numpy().astype('int').tolist()
        printlog(checkpoints=checkpoints)
        for epoch in range(self.max_epoch):
            # train model on full trainset
            model.train()
            loss_patch = []
            for iteration, (x, target) in enumerate(trainloader):
                x, target = x.cuda(), target.cuda()
                loss_patch.append(step(model, x, target, criterion, optimizer))
                if iteration in checkpoints:
                    loss = sum(loss_patch)/len(loss_patch)
                    printlog(epoch=epoch, iter=iteration, loss=loss, mode='train')
                    loss_patch = []
                # if epoch == 0 and iteration in [int(_r*len(trainloader)) for _r in (0.05, 0.1, 0.2, 0.5)]:  # quick check
                #     printlog(epoch=epoch, iter=iteration, checkpoint='quick check')
                #     torch.save(model.state_dict(), f'last.pth')
                if _test_mode:
                    break
            # validation on full valset
            model.eval()
            loss_patch = []
            with torch.no_grad():
                for iteration, (x, target) in enumerate(valloader):
                    x, target = x.cuda(), target.cuda()
                    loss_patch.append(inference(model, x, target, criterion))
                    if _test_mode:
                        break
            loss = sum(loss_patch)/len(loss_patch)
            printlog(epoch=epoch, iter=iteration, loss=loss, mode='validation')
            if _test_mode:
                break
            if loss < min_average_kldiv_test:
                min_average_kldiv_test = loss
                patience = self.max_training_patience
                torch.save(model.state_dict(), f'last.pth')
            else:
                patience -= 1
                if patience < 1:
                    break
        return model

    def performance(self, model, _test_mode=False):
        testloader = torch.utils.data.DataLoader(
            SALICON(root=self.root, train=False, transform=self.transform),
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )
        # test performance
        model.eval()
        softmax2d = nn.Softmax2d_()
        metrics_patch = {'CC': [nn.CorrelationCoefficientsLoss(), []]}
        with torch.no_grad():
            for iteration, (x, target) in enumerate(testloader):
                x, target = x.cuda(), target.cuda()
                y = softmax2d(model(x))
                logs = {'iter': iteration, 'mode': 'test'}
                for m in metrics_patch:
                    metrics_patch[m][1].append(metrics_patch[m][0](y, target))
                    v = sum(metrics_patch[m][1])/len(metrics_patch[m][1])
                    logs[m] = v
                printlog(**logs)
                if _test_mode:
                    break
        return metrics_patch
